import collections
import copy
from .Utils import _write_complex_object


class MetricData(object):
    """Data contract class for type MetricData.
    """

    ENVELOPE_TYPE_NAME = 'Microsoft.ApplicationInsights.Metric'

    DATA_TYPE_NAME = 'MetricData'

    _defaults = collections.OrderedDict([
        ('ver', 2),
        ('metrics', []),
        ('properties', {})
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
            'ver': 2,
            'metrics': [],
        }
        self._initialize()

    @property
    def ver(self):
        """The ver property.

        Returns:
            (int). the property value. (defaults to: 2)
        """
        return self._values['ver']

    @ver.setter
    def ver(self, value):
        """The ver property.

        Args:
            value (int). the property value.
        """
        self._values['ver'] = value

    @property
    def metrics(self):
        """The metrics property.

        Returns:
            (list). the property value. (defaults to: [])
        """
        return self._values['metrics']

    @metrics.setter
    def metrics(self, value):
        """The metrics property.

        Args:
            value (list). the property value.
        """
        self._values['metrics'] = value

    @property
    def properties(self):
        """The properties property.

        Returns:
            (hash). the property value. (defaults to: {})
        """
        if 'properties' in self._values:
            return self._values['properties']
        self._values['properties'] = copy.deepcopy(
            self._defaults['properties'])
        return self._values['properties']

    @properties.setter
    def properties(self, value):
        """The properties property.

        Args:
            value (hash). the property value.
        """
        if value == self._defaults['properties'] and 'properties' in self._values:
            del self._values['properties']
        else:
            self._values['properties'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



try:
    # Python 2.x
    from Queue import Queue, Empty
except ImportError:
    # Python 3.x
    from queue import Queue, Empty

try:
    from persistqueue import Empty as PersistEmpty
    from persistqueue import Queue as PersistQueue
except ImportError:
    PersistEmpty = Empty
    PersistQueue = None


class QueueBase(object):
    """The base class for all types of queues for use in conjunction with an implementation of :class:`SenderBase`.

    The queue will notify the sender that it needs to pick up items when it reaches :func:`max_queue_length`,
    or when the consumer calls :func:`flush`.
    """

    def __init__(self, sender, persistence_path=''):
        """Initializes a new instance of the class.

        Args:
            sender (:class:`SenderBase`) the sender object that will be used in conjunction with this queue.
            persistence_path (str) if set, persist the queue on disk into the provided directory.
        """
        if persistence_path and PersistQueue is None:
            raise ValueError(
                'persistence_path argument requires persist-queue dependency to be installed')
        elif persistence_path:
            self._queue = PersistQueue(persistence_path)
        else:
            self._queue = Queue()

        self._persistence_path = persistence_path
        self._max_queue_length = 500
        self._sender = sender
        if sender:
            self._sender.queue = self

    @property
    def max_queue_length(self):
        """The maximum number of items that will be held by the queue before the queue will call the :func:`flush`
        method.

        Args:
            value (int). the maximum queue length. The minimum allowed value is 1.

        Returns:
            int. the maximum queue size. (defaults to: 500)
        """
        return self._max_queue_length

    @max_queue_length.setter
    def max_queue_length(self, value):
        """The maximum number of items that will be held by the queue before the queue will call the :func:`flush`
        method.

        Args:
            value (int): the maximum queue length. The minimum allowed value is 1.

        Returns:
            int. the maximum queue size. (defaults to: 500)
        """
        if value < 1:
            value = 1
        self._max_queue_length = value

    @property
    def sender(self):
        """The sender that is associated with this queue that this queue will use to send data to the service.

        Returns:
            :class:`SenderBase`. the sender object.
        """
        return self._sender

    def put(self, item):
        """Adds the passed in item object to the queue and calls :func:`flush` if the size of the queue is larger
        than :func:`max_queue_length`. This method does nothing if the passed in item is None.

        Args:
            item (:class:`contracts.Envelope`) item the telemetry envelope object to send to the service.
        """
        if not item:
            return
        self._queue.put(item)
        if self._queue.qsize() >= self._max_queue_length:
            self.flush()

    def get(self):
        """Gets a single item from the queue and returns it. If the queue is empty, this method will return None.

        Returns:
            :class:`contracts.Envelope`. a telemetry envelope object or None if the queue is empty.
        """
        try:
            item = self._queue.get_nowait()
        except (Empty, PersistEmpty):
            return None

        if self._persistence_path:
            self._queue.task_done()

        return item

    def flush(self):
        """Flushes the current queue by notifying the {#sender}. This method needs to be overridden by a concrete
        implementations of the queue class.
        """
        pass



import collections

from django.conf import settings
import applicationinsights

ApplicationInsightsSettings = collections.namedtuple("ApplicationInsightsSettings", [
    "ikey",
    "channel_settings",
    "use_view_name",
    "record_view_arguments",
    "log_exceptions"])

ApplicationInsightsChannelSettings = collections.namedtuple("ApplicationInsightsChannelSettings", [
    "send_interval",
    "send_time",
    "endpoint"])


def load_settings():
    if hasattr(settings, "APPLICATION_INSIGHTS"):
        config = settings.APPLICATION_INSIGHTS
    elif hasattr(settings, "APPLICATIONINSIGHTS"):
        config = settings.APPLICATIONINSIGHTS
    else:
        config = {}

    if not isinstance(config, dict):
        config = {}

    return ApplicationInsightsSettings(
        ikey=config.get("ikey"),
        use_view_name=config.get("use_view_name", False),
        record_view_arguments=config.get("record_view_arguments", False),
        log_exceptions=config.get("log_exceptions", True),
        channel_settings=ApplicationInsightsChannelSettings(
            endpoint=config.get("endpoint"),
            send_interval=config.get("send_interval"),
            send_time=config.get("send_time")))


saved_clients = {}
saved_channels = {}


def create_client(aisettings=None):
    global saved_clients, saved_channels

    if aisettings is None:
        aisettings = load_settings()

    if aisettings in saved_clients:
        return saved_clients[aisettings]

    channel_settings = aisettings.channel_settings

    if channel_settings in saved_channels:
        channel = saved_channels[channel_settings]
    else:
        sender = applicationinsights.channel.AsynchronousSender(
            service_endpoint_uri=channel_settings.endpoint)

        if channel_settings.send_time is not None:
            sender.send_time = channel_settings.send_time
        if channel_settings.send_interval is not None:
            sender.send_interval = channel_settings.send_interval

        queue = applicationinsights.channel.AsynchronousQueue(sender)
        channel = applicationinsights.channel.TelemetryChannel(None, queue)
        saved_channels[channel_settings] = channel

    ikey = aisettings.ikey
    if ikey is None:
        return dummy_client("No ikey specified")

    client = applicationinsights.TelemetryClient(aisettings.ikey, channel)
    saved_clients[aisettings] = client
    return client


def dummy_client(reason):
    """Creates a dummy channel so even if we're not logging telemetry, we can still send
    along the real object to things that depend on it to exist"""

    sender = applicationinsights.channel.NullSender()
    queue = applicationinsights.channel.SynchronousQueue(sender)
    channel = applicationinsights.channel.TelemetryChannel(None, queue)
    return applicationinsights.TelemetryClient("00000000-0000-0000-0000-000000000000", channel)



import collections
import copy
from .Utils import _write_complex_object


class Envelope(object):
    """Data contract class for type Envelope.
    """
    _defaults = collections.OrderedDict([
        ('ver', 1),
        ('name', None),
        ('time', None),
        ('sampleRate', 100.0),
        ('seq', None),
        ('iKey', None),
        ('tags', {}),
        ('data', None)
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
            'ver': 1,
            'name': None,
            'time': None,
            'sampleRate': 100.0,
        }
        self._initialize()

    @property
    def ver(self):
        """The ver property.

        Returns:
            (int). the property value. (defaults to: 1)
        """
        if 'ver' in self._values:
            return self._values['ver']
        return self._defaults['ver']

    @ver.setter
    def ver(self, value):
        """The ver property.

        Args:
            value (int). the property value.
        """
        if value == self._defaults['ver'] and 'ver' in self._values:
            del self._values['ver']
        else:
            self._values['ver'] = value

    @property
    def name(self):
        """The name property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        return self._values['name']

    @name.setter
    def name(self, value):
        """The name property.

        Args:
            value (string). the property value.
        """
        self._values['name'] = value

    @property
    def time(self):
        """The time property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        return self._values['time']

    @time.setter
    def time(self, value):
        """The time property.

        Args:
            value (string). the property value.
        """
        self._values['time'] = value

    @property
    def sample_rate(self):
        """The sample_rate property.

        Returns:
            (float). the property value. (defaults to: 100.0)
        """
        if 'sampleRate' in self._values:
            return self._values['sampleRate']
        return self._defaults['sampleRate']

    @sample_rate.setter
    def sample_rate(self, value):
        """The sample_rate property.

        Args:
            value (float). the property value.
        """
        if value == self._defaults['sampleRate'] and 'sampleRate' in self._values:
            del self._values['sampleRate']
        else:
            self._values['sampleRate'] = value

    @property
    def seq(self):
        """The seq property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'seq' in self._values:
            return self._values['seq']
        return self._defaults['seq']

    @seq.setter
    def seq(self, value):
        """The seq property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['seq'] and 'seq' in self._values:
            del self._values['seq']
        else:
            self._values['seq'] = value

    @property
    def ikey(self):
        """The ikey property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'iKey' in self._values:
            return self._values['iKey']
        return self._defaults['iKey']

    @ikey.setter
    def ikey(self, value):
        """The ikey property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['iKey'] and 'iKey' in self._values:
            del self._values['iKey']
        else:
            self._values['iKey'] = value

    @property
    def tags(self):
        """The tags property.

        Returns:
            (hash). the property value. (defaults to: {})
        """
        if 'tags' in self._values:
            return self._values['tags']
        self._values['tags'] = copy.deepcopy(self._defaults['tags'])
        return self._values['tags']

    @tags.setter
    def tags(self, value):
        """The tags property.

        Args:
            value (hash). the property value.
        """
        if value == self._defaults['tags'] and 'tags' in self._values:
            del self._values['tags']
        else:
            self._values['tags'] = value

    @property
    def data(self):
        """The data property.

        Returns:
            (object). the property value. (defaults to: None)
        """
        if 'data' in self._values:
            return self._values['data']
        return self._defaults['data']

    @data.setter
    def data(self, value):
        """The data property.

        Args:
            value (object). the property value.
        """
        if value == self._defaults['data'] and 'data' in self._values:
            del self._values['data']
        else:
            self._values['data'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



from .SenderBase import SenderBase


class NullSender(SenderBase):
    """A sender class that does not send data.  Useful for debug mode, when
    telemetry may not be desired, with no changes to the object model.
    """

    def __init__(self, *args, **kwargs):
        super(NullSender, self).__init__("nil-endpoint", *args, **kwargs)

    def send(self, data):
        pass

    def start(self):
        pass

    def stop(self):
        pass



from .Domain import Domain
from .AvailabilityData import AvailabilityData
from .Base import Base
from .Application import Application
from .Device import Device
from .Location import Location
from .Operation import Operation
from .Session import Session
from .User import User
from .Cloud import Cloud
from .Internal import Internal
from .Data import Data
from .DataPointType import DataPointType
from .DataPoint import DataPoint
from .Envelope import Envelope
from .EventData import EventData
from .StackFrame import StackFrame
from .ExceptionDetails import ExceptionDetails
from .SeverityLevel import SeverityLevel
from .ExceptionData import ExceptionData
from .MessageData import MessageData
from .MetricData import MetricData
from .PageViewData import PageViewData
from .PageViewPerfData import PageViewPerfData
from .RemoteDependencyData import RemoteDependencyData
from .RequestData import RequestData



import collections
import copy
from .Utils import _write_complex_object


class Device(object):
    """Data contract class for type Device.
    """
    _defaults = collections.OrderedDict([
        ('ai.device.id', None),
        ('ai.device.locale', None),
        ('ai.device.model', None),
        ('ai.device.oemName', None),
        ('ai.device.osVersion', None),
        ('ai.device.type', None)
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
        }
        self._initialize()

    @property
    def id(self):
        """The id property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.device.id' in self._values:
            return self._values['ai.device.id']
        return self._defaults['ai.device.id']

    @id.setter
    def id(self, value):
        """The id property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.device.id'] and 'ai.device.id' in self._values:
            del self._values['ai.device.id']
        else:
            self._values['ai.device.id'] = value

    @property
    def locale(self):
        """The locale property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.device.locale' in self._values:
            return self._values['ai.device.locale']
        return self._defaults['ai.device.locale']

    @locale.setter
    def locale(self, value):
        """The locale property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.device.locale'] and 'ai.device.locale' in self._values:
            del self._values['ai.device.locale']
        else:
            self._values['ai.device.locale'] = value

    @property
    def model(self):
        """The model property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.device.model' in self._values:
            return self._values['ai.device.model']
        return self._defaults['ai.device.model']

    @model.setter
    def model(self, value):
        """The model property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.device.model'] and 'ai.device.model' in self._values:
            del self._values['ai.device.model']
        else:
            self._values['ai.device.model'] = value

    @property
    def oem_name(self):
        """The oem_name property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.device.oemName' in self._values:
            return self._values['ai.device.oemName']
        return self._defaults['ai.device.oemName']

    @oem_name.setter
    def oem_name(self, value):
        """The oem_name property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.device.oemName'] and 'ai.device.oemName' in self._values:
            del self._values['ai.device.oemName']
        else:
            self._values['ai.device.oemName'] = value

    @property
    def os_version(self):
        """The os_version property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.device.osVersion' in self._values:
            return self._values['ai.device.osVersion']
        return self._defaults['ai.device.osVersion']

    @os_version.setter
    def os_version(self, value):
        """The os_version property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.device.osVersion'] and 'ai.device.osVersion' in self._values:
            del self._values['ai.device.osVersion']
        else:
            self._values['ai.device.osVersion'] = value

    @property
    def type(self):
        """The type property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.device.type' in self._values:
            return self._values['ai.device.type']
        return self._defaults['ai.device.type']

    @type.setter
    def type(self, value):
        """The type property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.device.type'] and 'ai.device.type' in self._values:
            del self._values['ai.device.type']
        else:
            self._values['ai.device.type'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



import collections


def _write_complex_object(defaults, values):
    output = collections.OrderedDict()
    for key in defaults.keys():
        default = defaults[key]
        if key in values:
            value = values[key]
            if value == None:
                value = default
        elif default:
            value = default
        else:
            continue

        if isinstance(value, list):
            value_copy = []
            for item in value:
                if hasattr(item, 'write') and callable(getattr(item, 'write')):
                    value_copy.append(item.write())
                else:
                    value_copy.append(item)
            if len(value_copy) > 0:
                output[key] = value_copy
        elif isinstance(value, dict):
            value_copy = collections.OrderedDict()
            keys = sorted(value.keys())
            for item_key in keys:
                item_value = value[item_key]
                if hasattr(item_value, 'write') and callable(getattr(item_value, 'write')):
                    value_copy[item_key] = item_value.write()
                else:
                    value_copy[item_key] = item_value
            if len(value_copy) > 0:
                output[key] = value_copy
        elif hasattr(value, 'write') and callable(getattr(value, 'write')):
            value_copy = value.write()
            if len(value_copy) > 0:
                output[key] = value_copy
        else:
            value_copy = value
            output[key] = value_copy

    return output



from .AsynchronousQueue import AsynchronousQueue
from .AsynchronousSender import AsynchronousSender
from .SenderBase import SenderBase
from .QueueBase import QueueBase
from .SynchronousQueue import SynchronousQueue
from .SynchronousSender import SynchronousSender
from .TelemetryChannel import TelemetryChannel
from .TelemetryContext import TelemetryContext
from .NullSender import NullSender
from . import contracts



import collections
import copy
from .Utils import _write_complex_object
from .DataPointType import DataPointType


class DataPoint(object):
    """Data contract class for type DataPoint.
    """
    _defaults = collections.OrderedDict([
        ('ns', None),
        ('name', None),
        ('kind', DataPointType.measurement),
        ('value', None),
        ('count', None),
        ('min', None),
        ('max', None),
        ('stdDev', None)
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
            'name': None,
            'kind': DataPointType.measurement,
            'value': None,
        }
        self._initialize()

    @property
    def ns(self):
        """The ns property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ns' in self._values:
            return self._values['ns']
        return self._defaults['ns']

    @ns.setter
    def ns(self, value):
        """The ns property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ns'] and 'ns' in self._values:
            del self._values['ns']
        else:
            self._values['ns'] = value

    @property
    def name(self):
        """The name property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        return self._values['name']

    @name.setter
    def name(self, value):
        """The name property.

        Args:
            value (string). the property value.
        """
        self._values['name'] = value

    @property
    def kind(self):
        """The kind property.

        Returns:
            (:class:`DataPointType.measurement`). the property value. (defaults to: DataPointType.measurement)
        """
        if 'kind' in self._values:
            return self._values['kind']
        return self._defaults['kind']

    @kind.setter
    def kind(self, value):
        """The kind property.

        Args:
            value (:class:`DataPointType.measurement`). the property value.
        """
        if value == self._defaults['kind'] and 'kind' in self._values:
            del self._values['kind']
        else:
            self._values['kind'] = value

    @property
    def value(self):
        """The value property.

        Returns:
            (float). the property value. (defaults to: None)
        """
        return self._values['value']

    @value.setter
    def value(self, value):
        """The value property.

        Args:
            value (float). the property value.
        """
        self._values['value'] = value

    @property
    def count(self):
        """The count property.

        Returns:
            (int). the property value. (defaults to: None)
        """
        if 'count' in self._values:
            return self._values['count']
        return self._defaults['count']

    @count.setter
    def count(self, value):
        """The count property.

        Args:
            value (int). the property value.
        """
        if value == self._defaults['count'] and 'count' in self._values:
            del self._values['count']
        else:
            self._values['count'] = value

    @property
    def min(self):
        """The min property.

        Returns:
            (float). the property value. (defaults to: None)
        """
        if 'min' in self._values:
            return self._values['min']
        return self._defaults['min']

    @min.setter
    def min(self, value):
        """The min property.

        Args:
            value (float). the property value.
        """
        if value == self._defaults['min'] and 'min' in self._values:
            del self._values['min']
        else:
            self._values['min'] = value

    @property
    def max(self):
        """The max property.

        Returns:
            (float). the property value. (defaults to: None)
        """
        if 'max' in self._values:
            return self._values['max']
        return self._defaults['max']

    @max.setter
    def max(self, value):
        """The max property.

        Args:
            value (float). the property value.
        """
        if value == self._defaults['max'] and 'max' in self._values:
            del self._values['max']
        else:
            self._values['max'] = value

    @property
    def std_dev(self):
        """The std_dev property.

        Returns:
            (float). the property value. (defaults to: None)
        """
        if 'stdDev' in self._values:
            return self._values['stdDev']
        return self._defaults['stdDev']

    @std_dev.setter
    def std_dev(self, value):
        """The std_dev property.

        Args:
            value (float). the property value.
        """
        if value == self._defaults['stdDev'] and 'stdDev' in self._values:
            del self._values['stdDev']
        else:
            self._values['stdDev'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



import sys
from applicationinsights import TelemetryClient, channel

original_excepthook = None
telemetry_channel = None
enabled_instrumentation_keys = []


def enable(instrumentation_key, *args, **kwargs):
    """Enables the automatic collection of unhandled exceptions. Captured exceptions will be sent to the Application
    Insights service before being re-thrown. Multiple calls to this function with different instrumentation keys result
    in multiple instances being submitted, one for each key.

    .. code:: python

        from applicationinsights.exceptions import enable

        # set up exception capture
        enable('<YOUR INSTRUMENTATION KEY GOES HERE>')

        # raise an exception (this will be sent to the Application Insights service as an exception telemetry object)
        raise Exception('Boom!')

    Args:
        instrumentation_key (str). the instrumentation key to use while sending telemetry to the service.
    """
    if not instrumentation_key:
        raise Exception('Instrumentation key was required but not provided')
    global original_excepthook
    global telemetry_channel
    telemetry_channel = kwargs.get('telemetry_channel')
    if not original_excepthook:
        original_excepthook = sys.excepthook
        sys.excepthook = intercept_excepthook
    if instrumentation_key not in enabled_instrumentation_keys:
        enabled_instrumentation_keys.append(instrumentation_key)


def intercept_excepthook(type, value, traceback):
    client = TelemetryClient('temp_key', telemetry_channel)
    for instrumentation_key in enabled_instrumentation_keys:
        client.context.instrumentation_key = instrumentation_key
        client.track_exception(type, value, traceback)
    client.flush()
    original_excepthook(type, value, traceback)



import logging
import applicationinsights
from weakref import WeakValueDictionary
from applicationinsights.channel import AsynchronousSender, AsynchronousQueue
from applicationinsights.channel import SynchronousSender, SynchronousQueue
from applicationinsights.channel import TelemetryChannel

enabled_instrumentation_keys = WeakValueDictionary()


def enable(instrumentation_key, *args, **kwargs):
    """Enables the Application Insights logging handler for the root logger for the supplied instrumentation key.
    Multiple calls to this function with different instrumentation keys result in multiple handler instances.

    .. code:: python

        import logging
        from applicationinsights.logging import enable

        # set up logging
        enable('<YOUR INSTRUMENTATION KEY GOES HERE>')

        # log something (this will be sent to the Application Insights service as a trace)
        logging.info('This is a message')

        # logging shutdown will cause a flush of all un-sent telemetry items
        # alternatively set up an async channel via enable('<YOUR INSTRUMENTATION KEY GOES HERE>', async_=True)

    Args:
        instrumentation_key (str). the instrumentation key to use while sending telemetry to the service.

    Keyword Args:
        async_ (bool): Whether to use an async channel for the telemetry. Defaults to False.
        endpoint (str): The custom endpoint to which to send the telemetry. Defaults to None.
        level (Union[int, str]): The level to set for the logger. Defaults to INFO.

    Returns:
        :class:`ApplicationInsightsHandler`. the newly created or existing handler.
    """
    if not instrumentation_key:
        raise Exception('Instrumentation key was required but not provided')
    if instrumentation_key in enabled_instrumentation_keys:
        logging.getLogger().removeHandler(
            enabled_instrumentation_keys[instrumentation_key])
    async_ = kwargs.pop('async_', False)
    endpoint = kwargs.pop('endpoint', None)
    telemetry_channel = kwargs.get('telemetry_channel')
    if telemetry_channel and async_:
        raise Exception('Incompatible arguments async_ and telemetry_channel')
    if telemetry_channel and endpoint:
        raise Exception(
            'Incompatible arguments endpoint and telemetry_channel')
    if not telemetry_channel:
        if async_:
            sender, queue = AsynchronousSender, AsynchronousQueue
        else:
            sender, queue = SynchronousSender, SynchronousQueue
        kwargs['telemetry_channel'] = TelemetryChannel(
            queue=queue(sender(endpoint)))
    log_level = kwargs.pop('level', logging.INFO)
    handler = LoggingHandler(instrumentation_key, *args, **kwargs)
    handler.setLevel(log_level)
    enabled_instrumentation_keys[instrumentation_key] = handler
    logging.getLogger().addHandler(handler)
    return handler


class LoggingHandler(logging.Handler):
    """This class represents an integration point between Python's logging framework and the Application Insights
    service.

    Logging records are sent to the service either as simple Trace telemetry or as Exception telemetry (in the case
    of exception information being available).

    .. code:: python

        import logging
        from applicationinsights.logging import ApplicationInsightsHandler

        # set up logging
        handler = ApplicationInsightsHandler('<YOUR INSTRUMENTATION KEY GOES HERE>')
        logging.basicConfig(handlers=[ handler ], format='%(levelname)s: %(message)s', level=logging.DEBUG)

        # log something (this will be sent to the Application Insights service as a trace)
        logging.info('This is a message')

        # logging shutdown will cause a flush of all un-sent telemetry items
        # alternatively flush manually via handler.flush()
    """

    def __init__(self, instrumentation_key, *args, **kwargs):
        """
        Initialize a new instance of the class.

        Args:
            instrumentation_key (str). the instrumentation key to use while sending telemetry to the service.
        """
        if not instrumentation_key:
            raise Exception(
                'Instrumentation key was required but not provided')
        telemetry_channel = kwargs.get('telemetry_channel')
        if 'telemetry_channel' in kwargs:
            del kwargs['telemetry_channel']
        self.client = applicationinsights.TelemetryClient(
            instrumentation_key, telemetry_channel)
        super(LoggingHandler, self).__init__(*args, **kwargs)

    def flush(self):
        """Flushes the queued up telemetry to the service.
        """
        self.client.flush()
        return super(LoggingHandler, self).flush()

    def emit(self, record):
        """Emit a record.

        If a formatter is specified, it is used to format the record. If exception information is present, an Exception
        telemetry object is sent instead of a Trace telemetry object.

        Args:
            record (:class:`logging.LogRecord`). the record to format and send.
        """
        # the set of properties that will ride with the record
        properties = {
            'process': record.processName,
            'module': record.module,
            'fileName': record.filename,
            'lineNumber': record.lineno,
            'level': record.levelname,
        }

        # if we have exec_info, we will use it as an exception
        if record.exc_info:
            self.client.track_exception(
                *record.exc_info, properties=properties)
            return

        # if we don't simply format the message and send the trace
        formatted_message = self.format(record)
        self.client.track_trace(
            formatted_message, properties=properties, severity=record.levelname)



import collections
import copy
from .Utils import _write_complex_object


class AvailabilityData(object):
    """Data contract class for type AvailabilityData.
    """
    ENVELOPE_TYPE_NAME = 'Microsoft.ApplicationInsights.Availability'

    DATA_TYPE_NAME = 'AvailabilityData'

    _defaults = collections.OrderedDict([
        ('ver', 2),
        ('id', None),
        ('name', None),
        ('duration', None),
        ('success', None),
        ('runLocation', None),
        ('message', None),
        ('properties', {}),
        ('measurements', {})
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
            'ver': 2,
            'id': None,
            'name': None,
            'duration': None,
            'success': None,
        }
        self._initialize()

    @property
    def ver(self):
        """The ver property.

        Returns:
            (int). the property value. (defaults to: 2)
        """
        return self._values['ver']

    @ver.setter
    def ver(self, value):
        """The ver property.

        Args:
            value (int). the property value.
        """
        self._values['ver'] = value

    @property
    def id(self):
        """The id property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        return self._values['id']

    @id.setter
    def id(self, value):
        """The id property.

        Args:
            value (string). the property value.
        """
        self._values['id'] = value

    @property
    def name(self):
        """The name property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        return self._values['name']

    @name.setter
    def name(self, value):
        """The name property.

        Args:
            value (string). the property value.
        """
        self._values['name'] = value

    @property
    def duration(self):
        """The duration property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        return self._values['duration']

    @duration.setter
    def duration(self, value):
        """The duration property.

        Args:
            value (string). the property value.
        """
        self._values['duration'] = value

    @property
    def success(self):
        """The success property.

        Returns:
            (bool). the property value. (defaults to: None)
        """
        return self._values['success']

    @success.setter
    def success(self, value):
        """The success property.

        Args:
            value (bool). the property value.
        """
        self._values['success'] = value

    @property
    def run_location(self):
        """The run_location property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'runLocation' in self._values:
            return self._values['runLocation']
        return self._defaults['runLocation']

    @run_location.setter
    def run_location(self, value):
        """The run_location property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['runLocation'] and 'runLocation' in self._values:
            del self._values['runLocation']
        else:
            self._values['runLocation'] = value

    @property
    def message(self):
        """The message property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'message' in self._values:
            return self._values['message']
        return self._defaults['message']

    @message.setter
    def message(self, value):
        """The message property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['message'] and 'message' in self._values:
            del self._values['message']
        else:
            self._values['message'] = value

    @property
    def properties(self):
        """The properties property.

        Returns:
            (hash). the property value. (defaults to: {})
        """
        if 'properties' in self._values:
            return self._values['properties']
        self._values['properties'] = copy.deepcopy(
            self._defaults['properties'])
        return self._values['properties']

    @properties.setter
    def properties(self, value):
        """The properties property.

        Args:
            value (hash). the property value.
        """
        if value == self._defaults['properties'] and 'properties' in self._values:
            del self._values['properties']
        else:
            self._values['properties'] = value

    @property
    def measurements(self):
        """The measurements property.

        Returns:
            (hash). the property value. (defaults to: {})
        """
        if 'measurements' in self._values:
            return self._values['measurements']
        self._values['measurements'] = copy.deepcopy(
            self._defaults['measurements'])
        return self._values['measurements']

    @measurements.setter
    def measurements(self, value):
        """The measurements property.

        Args:
            value (hash). the property value.
        """
        if value == self._defaults['measurements'] and 'measurements' in self._values:
            del self._values['measurements']
        else:
            self._values['measurements'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



import datetime
import re
import uuid
import applicationinsights


class WSGIApplication(object):
    """ This class represents a WSGI wrapper that enables request telemetry for existing WSGI applications. The request
    telemetry will be sent to Application Insights service using the supplied instrumentation key.

    .. code:: python

            from applicationinsights.requests import WSGIApplication
            from paste.httpserver import serve
            from pyramid.response import Response
            from pyramid.view import view_config

            @view_config()
            def hello(request):
                return Response('Hello')

            if __name__ == '__main__':
                from pyramid.config import Configurator
                config = Configurator()
                config.scan()
                app = config.make_wsgi_app()

                # Enable Application Insights middleware
                app = WSGIApplication('<YOUR INSTRUMENTATION KEY GOES HERE>', app, common_properties={'service': 'hello_world_service'})

                serve(app, host='0.0.0.0')
    """

    def __init__(self, instrumentation_key, wsgi_application, *args, **kwargs):
        """
        Initialize a new instance of the class.

        Args:
            instrumentation_key (str). the instrumentation key to use while sending telemetry to the service.\n
            wsgi_application (func). the WSGI application that we're wrapping.
        """
        if not instrumentation_key:
            raise Exception(
                'Instrumentation key was required but not provided')
        if not wsgi_application:
            raise Exception('WSGI application was required but not provided')
        telemetry_channel = kwargs.pop('telemetry_channel', None)
        if not telemetry_channel:
            sender = applicationinsights.channel.AsynchronousSender()
            queue = applicationinsights.channel.AsynchronousQueue(sender)
            telemetry_channel = applicationinsights.channel.TelemetryChannel(
                None, queue)
        self.client = applicationinsights.TelemetryClient(
            instrumentation_key, telemetry_channel)
        self.client.context.device.type = "PC"
        self._wsgi_application = wsgi_application
        self._common_properties = kwargs.pop('common_properties', {})

    def flush(self):
        """Flushes the queued up telemetry to the service.
        """
        self.client.flush()

    def __call__(self, environ, start_response):
        """Callable implementation for WSGI middleware.

        Args:
            environ (dict). a dictionary containing all WSGI environment properties for this request.\n
            start_response (func). a function used to store the status, HTTP headers to be sent to the client and optional exception information.

        Returns:
            (obj). the response to send back to the client.
        """
        start_time = datetime.datetime.utcnow()
        name = environ.get('PATH_INFO') or '/'
        closure = {'status': '200 OK'}
        http_method = environ.get('REQUEST_METHOD', 'GET')

        self.client.context.operation.id = str(uuid.uuid4())
        # operation.parent_id ought to be the request id (not the operation id, but we don't have it yet)
        self.client.context.operation.name = http_method + ' ' + name

        try:
            self.client.context.location.ip = environ['HTTP_X_FORWARDED_FOR'].split(',')[-1].strip()
        except KeyError:
            try:
                self.client.context.location.ip = environ['REMOTE_ADDR']
            except KeyError:
                pass

        def status_interceptor(status_string, headers_array, exc_info=None):
            closure['status'] = status_string
            start_response(status_string, headers_array, exc_info)

        for part in self._wsgi_application(environ, status_interceptor):
            yield part

        success = True
        response_match = re.match(r'\s*(?P<code>\d+)', closure['status'])
        if response_match:
            response_code = response_match.group('code')
            if int(response_code) >= 400:
                success = False
        else:
            response_code = closure['status']
            success = False

        url = name
        query_string = environ.get('QUERY_STRING')
        if query_string:
            url += '?' + query_string

        scheme = environ.get('wsgi.url_scheme', 'http')
        host = environ.get('HTTP_HOST', environ.get('SERVER_NAME', 'unknown'))

        url = scheme + '://' + host + url

        end_time = datetime.datetime.utcnow()
        duration = int((end_time - start_time).total_seconds() * 1000)

        self.client.track_request(name, url, success, start_time.isoformat(
        ) + 'Z', duration, response_code, http_method, self._common_properties)



import collections
import copy
from .Utils import _write_complex_object


class Cloud(object):
    """Data contract class for type Cloud.
    """
    _defaults = collections.OrderedDict([
        ('ai.cloud.role', None),
        ('ai.cloud.roleInstance', None)
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
        }
        self._initialize()

    @property
    def role(self):
        """The role property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.cloud.role' in self._values:
            return self._values['ai.cloud.role']
        return self._defaults['ai.cloud.role']

    @role.setter
    def role(self, value):
        """The role property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.cloud.role'] and 'ai.cloud.role' in self._values:
            del self._values['ai.cloud.role']
        else:
            self._values['ai.cloud.role'] = value

    @property
    def role_instance(self):
        """The role_instance property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.cloud.roleInstance' in self._values:
            return self._values['ai.cloud.roleInstance']
        return self._defaults['ai.cloud.roleInstance']

    @role_instance.setter
    def role_instance(self, value):
        """The role_instance property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.cloud.roleInstance'] and 'ai.cloud.roleInstance' in self._values:
            del self._values['ai.cloud.roleInstance']
        else:
            self._values['ai.cloud.roleInstance'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



import datetime
import traceback
import sys
import uuid


from applicationinsights import channel

NULL_CONSTANT_STRING = 'Null'


class TelemetryClient(object):
    """The telemetry client used for sending all types of telemetry. It serves as the main entry point for
    interacting with the Application Insights service.
    """

    def __init__(self, instrumentation_key, telemetry_channel=None):
        """Initializes a new instance of the class.

        Args:
            instrumentation_key (str). the instrumentation key to use for this telemetry client.\n
            telemetry_channel (:class:`channel.TelemetryChannel`). the optional telemetry channel to be used instead of
                constructing a default one.
        """
        if instrumentation_key:
            if isinstance(instrumentation_key, channel.TelemetryChannel):
                telemetry_channel = instrumentation_key
                instrumentation_key = None
        else:
            raise Exception(
                'Instrumentation key was required but not provided')
        self._context = channel.TelemetryContext()
        self._context.instrumentation_key = instrumentation_key
        self._channel = telemetry_channel or channel.TelemetryChannel()
        self._telemetry_processors = []

    @property
    def context(self):
        """The context associated with this client. All data objects created by this client will be accompanied by
        this value.

        Returns:
            :class:`channel.TelemetryContext`. the context instance.
        """
        return self._context

    @property
    def channel(self):
        """The channel associated with this telemetry client. All data created by this client will be passed along with
        the :func:`context` object to :class:`channel.TelemetryChannel`'s :func:`write`.

        Returns:
            :class:`channel.TelemetryChannel`. the channel instance.
        """
        return self._channel

    def flush(self):
        """Flushes data in the queue. Data in the queue will be sent either immediately irrespective of what sender is
        being used.
        """
        self._channel.flush()

    def track_pageview(self, name, url, duration=0, properties=None, measurements=None):
        """Send information about the page viewed in the application (a web page for instance).

        Args:
            name (str). the name of the page that was viewed.\n
            url (str). the URL of the page that was viewed.\n
            duration (int). the duration of the page view in milliseconds. (defaults to: 0)\n
            properties (dict). the set of custom properties the client wants attached to this data item. (defaults to: None)\n
            measurements (dict). the set of custom measurements the client wants to attach to this data item. (defaults to: None)
        """
        data = channel.contracts.PageViewData()
        data.name = name or NULL_CONSTANT_STRING
        data.url = url
        data.duration = duration
        if properties:
            data.properties = properties
        if measurements:
            data.measurements = measurements

        self.track(data, self._context)

    def track_exception(self, type=None, value=None, tb=None, properties=None, measurements=None):
        """ Send information about a single exception that occurred in the application.

        Args:
            type (Type). the type of the exception that was thrown.\n
            value (:class:`Exception`). the exception that the client wants to send.\n
            tb (:class:`Traceback`). the traceback information as returned by :func:`sys.exc_info`.\n
            properties (dict). the set of custom properties the client wants attached to this data item. (defaults to: None)\n
            measurements (dict). the set of custom measurements the client wants to attach to this data item. (defaults to: None)
        """
        if not type or not value or not tb:
            type, value, tb = sys.exc_info()

        if not type or not value or not tb:
            try:
                raise Exception(NULL_CONSTANT_STRING)
            except:
                type, value, tb = sys.exc_info()

        details = channel.contracts.ExceptionDetails()
        details.id = 1
        details.outer_id = 0
        details.type_name = type.__name__
        details.message = str(value)
        details.has_full_stack = True
        counter = 0
        for tb_frame_file, tb_frame_line, tb_frame_function, tb_frame_text in traceback.extract_tb(tb):
            frame = channel.contracts.StackFrame()
            frame.assembly = 'Unknown'
            frame.file_name = tb_frame_file
            frame.level = counter
            frame.line = tb_frame_line
            frame.method = tb_frame_function
            details.parsed_stack.append(frame)
            counter += 1
        details.parsed_stack.reverse()

        data = channel.contracts.ExceptionData()
        data.handled_at = 'UserCode'
        data.exceptions.append(details)
        if properties:
            data.properties = properties
        if measurements:
            data.measurements = measurements
        self.track(data, self._context)

    def track_event(self, name, properties=None, measurements=None):
        """ Send information about a single event that has occurred in the context of the application.

        Args:
            name (str). the data to associate to this event.\n
            properties (dict). the set of custom properties the client wants attached to this data item. (defaults to: None)\n
            measurements (dict). the set of custom measurements the client wants to attach to this data item. (defaults to: None)
        """
        data = channel.contracts.EventData()
        data.name = name or NULL_CONSTANT_STRING
        if properties:
            data.properties = properties
        if measurements:
            data.measurements = measurements

        self.track(data, self._context)

    def track_metric(self, name, value, type=None, count=None, min=None, max=None, std_dev=None, properties=None):
        """Send information about a single metric data point that was captured for the application.

        Args:
            name (str). the name of the metric that was captured.\n
            value (float). the value of the metric that was captured.\n
            type (:class:`channel.contracts.DataPointType`). the type of the metric. (defaults to: :func:`channel.contracts.DataPointType.aggregation`)\n
            count (int). the number of metrics that were aggregated into this data point. (defaults to: None)\n
            min (float). the minimum of all metrics collected that were aggregated into this data point. (defaults to: None)\n
            max (float). the maximum of all metrics collected that were aggregated into this data point. (defaults to: None)\n
            std_dev (float). the standard deviation of all metrics collected that were aggregated into this data point. (defaults to: None)\n
            properties (dict). the set of custom properties the client wants attached to this data item. (defaults to: None)
        """
        dataPoint = channel.contracts.DataPoint()
        dataPoint.name = name or NULL_CONSTANT_STRING
        dataPoint.value = value or 0
        dataPoint.kind = type or channel.contracts.DataPointType.aggregation
        dataPoint.count = count
        dataPoint.min = min
        dataPoint.max = max
        dataPoint.std_dev = std_dev

        data = channel.contracts.MetricData()
        data.metrics.append(dataPoint)
        if properties:
            data.properties = properties

        self.track(data, self._context)

    def track_trace(self, name, properties=None, severity=None):
        """Sends a single trace statement.

        Args:
            name (str). the trace statement.\n
            properties (dict). the set of custom properties the client wants attached to this data item. (defaults to: None)\n
            severity (str). the severity level of this trace, one of DEBUG, INFO, WARNING, ERROR, CRITICAL
        """
        data = channel.contracts.MessageData()
        data.message = name or NULL_CONSTANT_STRING
        if properties:
            data.properties = properties
        if severity is not None:
            data.severity_level = channel.contracts.MessageData.PYTHON_LOGGING_LEVELS.get(
                severity)

        self.track(data, self._context)

    def track_request(self, name, url, success, start_time=None, duration=None, response_code=None, http_method=None, properties=None, measurements=None, request_id=None):
        """Sends a single request that was captured for the application.

        Args:
            name (str). the name for this request. All requests with the same name will be grouped together.\n
            url (str). the actual URL for this request (to show in individual request instances).\n
            success (bool). true if the request ended in success, false otherwise.\n
            start_time (str). the start time of the request. The value should look the same as the one returned by :func:`datetime.isoformat()` (defaults to: None)\n
            duration (int). the number of milliseconds that this request lasted. (defaults to: None)\n
            response_code (str). the response code that this request returned. (defaults to: None)\n
            http_method (str). the HTTP method that triggered this request. (defaults to: None)\n
            properties (dict). the set of custom properties the client wants attached to this data item. (defaults to: None)\n
            measurements (dict). the set of custom measurements the client wants to attach to this data item. (defaults to: None)\n
            request_id (str). the id for this request. If None, a new uuid will be generated. (defaults to: None)
        """
        data = channel.contracts.RequestData()
        data.id = request_id or str(uuid.uuid4())
        data.name = name
        data.url = url
        data.success = success
        data.start_time = start_time or datetime.datetime.utcnow().isoformat() + 'Z'
        data.duration = self.__ms_to_duration(duration)
        data.response_code = str(response_code) or '200'
        data.http_method = http_method or 'GET'
        if properties:
            data.properties = properties
        if measurements:
            data.measurements = measurements

        self.track(data, self._context)

    def track_dependency(self, name, data, type=None, target=None, duration=None, success=None, result_code=None, properties=None, measurements=None, dependency_id=None):
        """Sends a single dependency telemetry that was captured for the application.

        Args:
            name (str). the name of the command initiated with this dependency call. Low cardinality value. Examples are stored procedure name and URL path template.\n
            data (str). the command initiated by this dependency call. Examples are SQL statement and HTTP URL with all query parameters.\n
            type (str). the dependency type name. Low cardinality value for logical grouping of dependencies and interpretation of other fields like commandName and resultCode. Examples are SQL, Azure table, and HTTP. (default to: None)\n
            target (str). the target site of a dependency call. Examples are server name, host address. (default to: None)\n
            duration (int). the number of milliseconds that this dependency call lasted. (defaults to: None)\n
            success (bool). true if the dependency call ended in success, false otherwise. (defaults to: None)\n
            result_code (str). the result code of a dependency call. Examples are SQL error code and HTTP status code. (defaults to: None)\n
            properties (dict). the set of custom properties the client wants attached to this data item. (defaults to: None)\n
            measurements (dict). the set of custom measurements the client wants to attach to this data item. (defaults to: None)\n
            id (str). the id for this dependency call. If None, a new uuid will be generated. (defaults to: None)
        """
        dependency_data = channel.contracts.RemoteDependencyData()
        dependency_data.id = dependency_id or str(uuid.uuid4())
        dependency_data.name = name
        dependency_data.data = data
        dependency_data.type = type
        dependency_data.target = target
        dependency_data.duration = self.__ms_to_duration(duration)
        dependency_data.success = success
        dependency_data.result_code = str(result_code) or '200'
        if properties:
            dependency_data.properties = properties
        if measurements:
            dependency_data.measurements = measurements

        self.track(dependency_data, self._context)

    def track_availability(self, name, duration, success, run_location, message=None, properties=None, measurements=None, availability_id=None):
        """Sends a single availability telemetry that was captured for the application.

        Args:
            name (str). the name of the command initiated with this dependency call. Low cardinality value. Examples are stored procedure name and URL path template.\n
            duration (int). the number of milliseconds that this dependency call lasted.\n
            success (bool). true if the dependency call ended in success, false otherwise.\n
            run_location (str). the name of the location where the test has been executed.\n
            message (str). the message for this availability telemetry. (defaults to: None)\n
            properties (dict). the set of custom properties the client wants attached to this data item. (defaults to: None)\n
            measurements (dict). the set of custom measurements the client wants to attach to this data item. (defaults to: None)\n
            id (str). the id for this dependency call. If None, a new uuid will be generated. (defaults to: None)
        """
        availability_data = channel.contracts.AvailabilityData()
        availability_data.id = availability_id or str(uuid.uuid4())
        availability_data.name = name
        availability_data.duration = self.__ms_to_duration(duration)
        availability_data.success = success
        availability_data.run_location = run_location
        availability_data.message = message
        if properties:
            availability_data.properties = properties
        if measurements:
            availability_data.measurements = measurements

        self.track(availability_data, self._context)

    def track(self, data, context):
        if self.run_telemetry_processors(data, context):
            self.channel.write(data, context)

    def add_telemetry_processor(self, telemetry_processor):
        """Adds telemetry processor to the collection. Telemetry processors will be called one by one
           before telemetry item is pushed for sending and in the order they were added.

        Args:
            telemetry_processor (function). Takes a telemetry item, and context object and returns boolean 
                                            that determines if the event is passed to the server (False = Filtered)
        """
        if telemetry_processor is None:
            raise TypeError('telemetry_processor cannot be None.')

        self._telemetry_processors.insert(0, telemetry_processor)

    def run_telemetry_processors(self, data, context):
        allow_data_through = True

        try:
            for processor in self._telemetry_processors:
                if processor(data, context) == False:
                    allow_data_through = False
                    break
        except:
            allow_data_through = True
        return allow_data_through

    @staticmethod
    def __ms_to_duration(duration_ms):
        local_duration = duration_ms or 0
        duration_parts = []
        for multiplier in [1000, 60, 60, 24]:
            duration_parts.append(local_duration % multiplier)
            local_duration //= multiplier

        duration_parts.reverse()
        duration = '%02d:%02d:%02d.%03d' % (
            duration_parts[0], duration_parts[1], duration_parts[2], duration_parts[3])
        if local_duration:
            duration = '%d.%s' % (local_duration, duration)

        return duration




import datetime
import inspect
import sys
import time
import uuid

from django.http import Http404

import applicationinsights
from applicationinsights.channel import contracts
from . import common

try:
    basestring     # Python 2
except NameError:  # Python 3
    basestring = (str, )

# Pick a function to measure time; starting with 3.3, time.monotonic is available.
try:
    TIME_FUNC = time.monotonic
except AttributeError:
    TIME_FUNC = time.time


class ApplicationInsightsMiddleware(object):
    """This class is a Django middleware that automatically enables request and exception telemetry.  Django versions
    1.7 and newer are supported.

    To enable, add this class to your settings.py file in MIDDLEWARE_CLASSES (pre-1.10) or MIDDLEWARE (1.10 and newer):

    .. code:: python

        # If on Django < 1.10
        MIDDLEWARE_CLASSES = [
            # ... or whatever is below for you ...
            'django.middleware.security.SecurityMiddleware',
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.middleware.common.CommonMiddleware',
            'django.middleware.csrf.CsrfViewMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
            'django.contrib.messages.middleware.MessageMiddleware',
            'django.middleware.clickjacking.XFrameOptionsMiddleware',
            # ... or whatever is above for you ...
            'applicationinsights.django.ApplicationInsightsMiddleware',   # Add this middleware to the end
        ]

        # If on Django >= 1.10
        MIDDLEWARE = [
            # ... or whatever is below for you ...
            'django.middleware.security.SecurityMiddleware',
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.middleware.common.CommonMiddleware',
            'django.middleware.csrf.CsrfViewMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            'django.contrib.messages.middleware.MessageMiddleware',
            'django.middleware.clickjacking.XFrameOptionsMiddleware',
            # ... or whatever is above for you ...
            'applicationinsights.django.ApplicationInsightsMiddleware',   # Add this middleware to the end
        ]

    And then, add the following to your settings.py file:

    .. code:: python

        APPLICATION_INSIGHTS = {
            # (required) Your Application Insights instrumentation key
            'ikey': "00000000-0000-0000-0000-000000000000",

            # (optional) By default, request names are logged as the request method
            # and relative path of the URL.  To log the fully-qualified view names
            # instead, set this to True.  Defaults to False.
            'use_view_name': True,

            # (optional) To log arguments passed into the views as custom properties,
            # set this to True.  Defaults to False.
            'record_view_arguments': True,

            # (optional) Exceptions are logged by default, to disable, set this to False.
            'log_exceptions': False,

            # (optional) Events are submitted to Application Insights asynchronously.
            # send_interval specifies how often the queue is checked for items to submit.
            # send_time specifies how long the sender waits for new input before recycling
            # the background thread.
            'send_interval': 1.0, # Check every second
            'send_time': 3.0, # Wait up to 3 seconds for an event

            # (optional, uncommon) If you must send to an endpoint other than the
            # default endpoint, specify it here:
            'endpoint': "https://dc.services.visualstudio.com/v2/track",
        }

    Once these are in place, each request will have an `appinsights` object placed on it.
    This object will have the following properties:

    * `client`: This is an instance of the :class:`applicationinsights.TelemetryClient` type, which will
      submit telemetry to the same instrumentation key, and will parent each telemetry item to the current
      request.
    * `request`: This is the :class:`applicationinsights.channel.contracts.RequestData` instance for the
      current request.  You can modify properties on this object during the handling of the current request.
      It will be submitted when the request has finished.
    * `context`: This is the :class:`applicationinsights.channel.TelemetryContext` object for the current
      ApplicationInsights sender.

    These properties will be present even when `DEBUG` is `True`, but it may not submit telemetry unless
    `debug_ikey` is set in `APPLICATION_INSIGHTS`, above.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

        # Get configuration
        self._settings = common.load_settings()
        self._client = common.create_client(self._settings)

    # Pre-1.10 handler
    def process_request(self, request):
        # Populate context object onto request
        addon = RequestAddon(self._client)
        data = addon.request
        context = addon.context
        request.appinsights = addon

        # Basic request properties
        data.start_time = datetime.datetime.utcnow().isoformat() + "Z"
        data.http_method = request.method
        data.url = request.build_absolute_uri()
        data.name = "%s %s" % (request.method, request.path)
        context.operation.name = data.name
        context.operation.id = data.id
        context.location.ip = request.META.get('REMOTE_ADDR', '')
        context.user.user_agent = request.META.get('HTTP_USER_AGENT', '')

        # User
        if hasattr(request, 'user'):
            if request.user is not None and not request.user.is_anonymous and request.user.is_authenticated:
                context.user.account_id = request.user.get_short_name()

        # Run and time the request
        addon.start_stopwatch()
        return None

    # Pre-1.10 handler
    def process_response(self, request, response):
        if hasattr(request, 'appinsights'):
            addon = request.appinsights
            duration = addon.measure_duration()

            data = addon.request
            context = addon.context

            # Fill in data from the response
            data.duration = addon.measure_duration()
            data.response_code = response.status_code
            data.success = response.status_code < 400 or response.status_code == 401

            # Submit and return
            self._client.channel.write(data, context)

        return response

    # 1.10 and up...
    def __call__(self, request):
        self.process_request(request)
        response = self.get_response(request)
        self.process_response(request, response)
        return response

    def process_view(self, request, view_func, view_args, view_kwargs):
        if not hasattr(request, "appinsights"):
            return None

        data = request.appinsights.request
        context = request.appinsights.context

        # Operation name is the method + url by default (set in __call__),
        # If use_view_name is set, then we'll look up the name of the view.
        if self._settings.use_view_name:
            mod = inspect.getmodule(view_func)
            if hasattr(view_func, "__name__"):
                name = view_func.__name__
            elif hasattr(view_func, "__class__") and hasattr(view_func.__class__, "__name__"):
                name = view_func.__class__.__name__
            else:
                name = "<unknown>"

            if mod:
                opname = "%s %s.%s" % (data.http_method, mod.__name__, name)
            else:
                opname = "%s %s" % (data.http_method, name)
            data.name = opname
            context.operation.name = opname

        # Populate the properties with view arguments
        if self._settings.record_view_arguments:
            for i, arg in enumerate(view_args):
                data.properties['view_arg_' + str(i)] = arg_to_str(arg)

            for k, v in view_kwargs.items():
                data.properties['view_arg_' + k] = arg_to_str(v)

        return None

    def process_exception(self, request, exception):
        if not self._settings.log_exceptions:
            return None

        if type(exception) is Http404:
            return None

        _, _, tb = sys.exc_info()
        if tb is None or exception is None:
            # No actual traceback or exception info, don't bother logging.
            return None

        client = applicationinsights.TelemetryClient(
            self._client.context.instrumentation_key, self._client.channel)
        if hasattr(request, 'appinsights'):
            client.context.operation.parent_id = request.appinsights.request.id

        client.track_exception(type(exception), exception, tb)

        return None

    def process_template_response(self, request, response):
        if hasattr(request, 'appinsights') and hasattr(response, 'template_name'):
            data = request.appinsights.request
            data.properties['template_name'] = response.template_name

        return response


class RequestAddon(object):
    def __init__(self, client):
        self._baseclient = client
        self._client = None
        self.request = contracts.RequestData()
        self.request.id = str(uuid.uuid4())
        self.context = applicationinsights.channel.TelemetryContext()
        self.context.instrumentation_key = client.context.instrumentation_key
        self.context.operation.id = self.request.id
        self._process_start_time = None

    @property
    def client(self):
        if self._client is None:
            # Create a client that submits telemetry parented to the request.
            self._client = applicationinsights.TelemetryClient(
                self.context.instrumentation_key, self._baseclient.channel)
            self._client.context.operation.parent_id = self.context.operation.id

        return self._client

    def start_stopwatch(self):
        self._process_start_time = TIME_FUNC()

    def measure_duration(self):
        end_time = TIME_FUNC()
        return ms_to_duration(int((end_time - self._process_start_time) * 1000))


def ms_to_duration(n):
    duration_parts = []
    for multiplier in [1000, 60, 60, 24]:
        duration_parts.append(n % multiplier)
        n //= multiplier

    duration_parts.reverse()
    duration = "%02d:%02d:%02d.%03d" % tuple(duration_parts)
    if n:
        duration = "%d.%s" % (n, duration)

    return duration


def arg_to_str(arg):
    if isinstance(arg, basestring):
        return arg
    if isinstance(arg, int):
        return str(arg)
    return repr(arg)



import json

try:
    # Python 2.x
    import urllib2 as HTTPClient
    from urllib2 import HTTPError
except ImportError:
    # Python 3.x
    import urllib.request as HTTPClient
    from urllib.error import HTTPError

DEFAULT_ENDPOINT_URL = 'https://dc.services.visualstudio.com/v2/track'


class SenderBase(object):
    """The base class for all types of senders for use in conjunction with an implementation of :class:`QueueBase`.

    The queue will notify the sender that it needs to pick up items. The concrete sender implementation will
    listen to these notifications and will pull items from the queue getting at most :func:`send_buffer_size` items.
    It will then call :func:`send` using the list of items pulled from the queue.
    """

    def __init__(self, service_endpoint_uri):
        """Initializes a new instance of the class.

        Args:
            service_endpoint_uri (str) the address of the service to send telemetry data to.
        """
        self._service_endpoint_uri = service_endpoint_uri
        self._queue = None
        self._send_buffer_size = 100
        self._timeout = 10

    @property
    def service_endpoint_uri(self):
        """The HTTP or HTTPS endpoint that this sender will send data to.

        Args:
            value (str). the service endpoint URI.

        Returns:
            str. the service endpoint URI.
        """
        return self._service_endpoint_uri

    @service_endpoint_uri.setter
    def service_endpoint_uri(self, value):
        """The service endpoint URI where this sender will send data to.

        Args:
            value (str). the service endpoint URI.

        Returns:
            str. the service endpoint URI.
        """
        self._service_endpoint_uri = value

    @property
    def queue(self):
        """The queue that this sender is draining. While :class:`SenderBase` doesn't implement any means of doing
        so, derivations of this class do.

        Args:
            value (:class:`QueueBase`). the queue instance that this sender is draining.

        Returns:
            :class:`QueueBase`. the queue instance that this sender is draining.
        """
        return self._queue

    @property
    def send_timeout(self):
        """Time in seconds that the sender should wait before giving up."""
        return self._timeout

    @send_timeout.setter
    def send_timeout(self, seconds):
        """Configures the timeout in seconds the sender waits for a response for the server.

        Args:
            seconds(float). Timeout in seconds.
        """

        self._timeout = seconds

    @queue.setter
    def queue(self, value):
        """The queue that this sender is draining. While :class:`SenderBase` doesn't implement any means of doing
        so, derivations of this class do.

        Args:
            value (:class:`QueueBase`). the queue instance that this sender is draining.

        Returns:
            :class:`QueueBase`. the queue instance that this sender is draining.
        """
        self._queue = value

    @property
    def send_buffer_size(self):
        """The buffer size for a single batch of telemetry. This is the maximum number of items in a single service
        request that this sender is going to send.

        Args:
            value (int). the maximum number of items in a telemetry batch.

        Returns:
            int. the maximum number of items in a telemetry batch.
        """
        return self._send_buffer_size

    @send_buffer_size.setter
    def send_buffer_size(self, value):
        """The buffer size for a single batch of telemetry. This is the maximum number of items in a single service
        request that this sender is going to send.

        Args:
            value (int). the maximum number of items in a telemetry batch.

        Returns:
            int. the maximum number of items in a telemetry batch.
        """
        if value < 1:
            value = 1
        self._send_buffer_size = value

    def send(self, data_to_send):
        """ Immediately sends the data passed in to :func:`service_endpoint_uri`. If the service request fails, the
        passed in items are pushed back to the :func:`queue`.

        Args:
            data_to_send (Array): an array of :class:`contracts.Envelope` objects to send to the service.
        """
        request_payload = json.dumps([a.write() for a in data_to_send])

        request = HTTPClient.Request(self._service_endpoint_uri, bytearray(
            request_payload, 'utf-8'), {'Accept': 'application/json', 'Content-Type': 'application/json; charset=utf-8'})
        try:
            response = HTTPClient.urlopen(request, timeout=self._timeout)
            status_code = response.getcode()
            if 200 <= status_code < 300:
                return
        except HTTPError as e:
            if e.getcode() == 400:
                return
        except Exception as e:
            pass

        # Add our unsent data back on to the queue
        for data in data_to_send:
            self._queue.put(data)



import collections
import copy
from .Utils import _write_complex_object


class Session(object):
    """Data contract class for type Session.
    """
    _defaults = collections.OrderedDict([
        ('ai.session.id', None),
        ('ai.session.isFirst', None)
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
        }
        self._initialize()

    @property
    def id(self):
        """The id property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.session.id' in self._values:
            return self._values['ai.session.id']
        return self._defaults['ai.session.id']

    @id.setter
    def id(self, value):
        """The id property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.session.id'] and 'ai.session.id' in self._values:
            del self._values['ai.session.id']
        else:
            self._values['ai.session.id'] = value

    @property
    def is_first(self):
        """The is_first property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.session.isFirst' in self._values:
            return self._values['ai.session.isFirst']
        return self._defaults['ai.session.isFirst']

    @is_first.setter
    def is_first(self, value):
        """The is_first property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.session.isFirst'] and 'ai.session.isFirst' in self._values:
            del self._values['ai.session.isFirst']
        else:
            self._values['ai.session.isFirst'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



import collections
import copy
from .Utils import _write_complex_object


class StackFrame(object):
    """Data contract class for type StackFrame.
    """
    _defaults = collections.OrderedDict([
        ('level', None),
        ('method', None),
        ('assembly', None),
        ('fileName', None),
        ('line', None)
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
            'level': None,
            'method': None,
        }
        self._initialize()

    @property
    def level(self):
        """The level property.

        Returns:
            (int). the property value. (defaults to: None)
        """
        return self._values['level']

    @level.setter
    def level(self, value):
        """The level property.

        Args:
            value (int). the property value.
        """
        self._values['level'] = value

    @property
    def method(self):
        """The method property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        return self._values['method']

    @method.setter
    def method(self, value):
        """The method property.

        Args:
            value (string). the property value.
        """
        self._values['method'] = value

    @property
    def assembly(self):
        """The assembly property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'assembly' in self._values:
            return self._values['assembly']
        return self._defaults['assembly']

    @assembly.setter
    def assembly(self, value):
        """The assembly property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['assembly'] and 'assembly' in self._values:
            del self._values['assembly']
        else:
            self._values['assembly'] = value

    @property
    def file_name(self):
        """The file_name property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'fileName' in self._values:
            return self._values['fileName']
        return self._defaults['fileName']

    @file_name.setter
    def file_name(self, value):
        """The file_name property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['fileName'] and 'fileName' in self._values:
            del self._values['fileName']
        else:
            self._values['fileName'] = value

    @property
    def line(self):
        """The line property.

        Returns:
            (int). the property value. (defaults to: None)
        """
        if 'line' in self._values:
            return self._values['line']
        return self._defaults['line']

    @line.setter
    def line(self, value):
        """The line property.

        Args:
            value (int). the property value.
        """
        if value == self._defaults['line'] and 'line' in self._values:
            del self._values['line']
        else:
            self._values['line'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



import platform
import locale

from applicationinsights.channel import contracts

# save off whatever is currently there
existing_device_initialize = contracts.Device._initialize


def device_initialize(self):
    """ The device initializer used to assign special properties to all device context objects"""
    existing_device_initialize(self)
    self.type = 'Other'
    self.id = platform.node()
    self.os_version = platform.version()
    self.locale = locale.getdefaultlocale()[0]


# assign the device context initializer
contracts.Device._initialize = device_initialize


class TelemetryContext(object):
    """Represents the context for sending telemetry to the Application Insights service.

    .. code:: python

        context = TelemetryContext()
        context.instrumentation_key = '<YOUR INSTRUMENTATION KEY GOES HERE>'
        context.application.ver = '1.2.3'
        context.device.id = 'My current device'
        context.device.oem_name = 'Asus'
        context.device.model = 'X31A'
        context.device.type = "Other"
        context.user.id = 'santa@northpole.net'
        track_trace('My trace with context')
    """

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self.instrumentation_key = None
        self.device = contracts.Device()
        self.cloud = contracts.Cloud()
        self.application = contracts.Application()
        self.user = contracts.User()
        self.session = contracts.Session()
        self.operation = contracts.Operation()
        self.location = contracts.Location()
        self._properties = {}

    @property
    def properties(self):
        """The property context. This contains free-form properties that you can add to your telemetry.

        Returns:
            (dict). the context object.
        """
        return self._properties



import datetime
import sys

from .SynchronousQueue import SynchronousQueue
from .SynchronousSender import SynchronousSender
from .TelemetryContext import TelemetryContext
from applicationinsights.channel import contracts

platform_moniker = 'py2'
if sys.version_info >= (3, 0):
    platform_moniker = 'py3'

# set up internal context
internal_context = contracts.Internal()
internal_context.sdk_version = platform_moniker + ':0.12.0'


class TelemetryChannel(object):
    """The telemetry channel is responsible for constructing a :class:`contracts.Envelope` object from the passed in
    data and specified telemetry context.

    .. code:: python

        from application_insights.channel import TelemetryChannel, contracts
        channel = TelemetryChannel()
        event = contracts.EventData()
        event.name = 'My event'
        channel.write(event)
    """

    def __init__(self, context=None, queue=None):
        """Initializes a new instance of the class.

        Args:
            context (:class:`TelemetryContext') the telemetry context to use when sending telemetry data.\n
            queue (:class:`QueueBase`) the queue to enqueue the resulting :class:`contracts.Envelope` to.
        """
        self._context = context or TelemetryContext()
        self._queue = queue or SynchronousQueue(SynchronousSender())

    @property
    def context(self):
        """The context associated with this channel. All :class:`contracts.Envelope` objects created by this channel
        will use this value if it's present or if none is specified as part of the :func:`write` call.

        Returns:
            (:class:`TelemetryContext`). the context instance (defaults to: TelemetryContext())
        """
        return self._context

    @property
    def queue(self):
        """The queue associated with this channel. All :class:`contracts.Envelope` objects created by this channel
        will be pushed to this queue.

        Returns:
            (:class:`QueueBase`). the queue instance (defaults to: SynchronousQueue())
        """
        return self._queue

    @property
    def sender(self):
        """The sender associated with this channel. This instance will be used to transmit telemetry to the service.

        Returns:
            (:class:`SenderBase`). the sender instance (defaults to: SynchronousSender())
        """
        return self._queue.sender

    def flush(self):
        """Flushes the enqueued data by calling :func:`flush` on :func:`queue`.
        """
        self._queue.flush()

    def write(self, data, context=None):
        """Enqueues the passed in data to the :func:`queue`. If the caller specifies a context as well, it will
        take precedence over the instance in :func:`context`.

        Args:
            data (object). data the telemetry data to send. This will be wrapped in an :class:`contracts.Envelope`
                before being enqueued to the :func:`queue`.
            context (:class:`TelemetryContext`). context the override context to use when constructing the
                :class:`contracts.Envelope`.
        """
        local_context = context or self._context
        if not local_context:
            raise Exception('Context was required but not provided')

        if not data:
            raise Exception('Data was required but not provided')

        envelope = contracts.Envelope()
        envelope.name = data.ENVELOPE_TYPE_NAME
        envelope.time = datetime.datetime.utcnow().isoformat() + 'Z'
        envelope.ikey = local_context.instrumentation_key
        tags = envelope.tags
        for prop_context in [self._context, context]:
            if not prop_context:
                continue
            for key, value in self._write_tags(prop_context):
                tags[key] = value
        envelope.data = contracts.Data()
        envelope.data.base_type = data.DATA_TYPE_NAME
        for prop_context in [context, self._context]:
            if not prop_context:
                continue
            if hasattr(data, 'properties') and prop_context.properties:
                properties = data.properties
                for key in prop_context.properties:
                    if key not in properties:
                        properties[key] = prop_context.properties[key]
        envelope.data.base_data = data

        self._queue.put(envelope)

    def _write_tags(self, context):
        for item in [internal_context,
                     context.device, context.cloud, context.application, context.user,
                     context.session, context.location, context.operation]:
            if not item:
                continue
            for pair in item.write().items():
                yield pair



import collections
import copy
from .Utils import _write_complex_object


class User(object):
    """Data contract class for type User.
    """
    _defaults = collections.OrderedDict([
        ('ai.user.accountId', None),
        ('ai.user.id', None),
        ('ai.user.authUserId', None)
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
        }
        self._initialize()

    @property
    def account_id(self):
        """The account_id property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.user.accountId' in self._values:
            return self._values['ai.user.accountId']
        return self._defaults['ai.user.accountId']

    @account_id.setter
    def account_id(self, value):
        """The account_id property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.user.accountId'] and 'ai.user.accountId' in self._values:
            del self._values['ai.user.accountId']
        else:
            self._values['ai.user.accountId'] = value

    @property
    def id(self):
        """The id property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.user.id' in self._values:
            return self._values['ai.user.id']
        return self._defaults['ai.user.id']

    @id.setter
    def id(self, value):
        """The id property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.user.id'] and 'ai.user.id' in self._values:
            del self._values['ai.user.id']
        else:
            self._values['ai.user.id'] = value

    @property
    def auth_user_id(self):
        """The auth_user_id property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.user.authUserId' in self._values:
            return self._values['ai.user.authUserId']
        return self._defaults['ai.user.authUserId']

    @auth_user_id.setter
    def auth_user_id(self, value):
        """The auth_user_id property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.user.authUserId'] and 'ai.user.authUserId' in self._values:
            del self._values['ai.user.authUserId']
        else:
            self._values['ai.user.authUserId'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



from .SenderBase import SenderBase, DEFAULT_ENDPOINT_URL
from threading import Lock, Thread


class AsynchronousSender(SenderBase):
    """An asynchronous sender that works in conjunction with the :class:`AsynchronousQueue`. The sender object will
    start a worker thread that will pull items from the :func:`queue`. The thread will be created when the client
    calls :func:`start` and will check for queue items every :func:`send_interval` seconds. The worker thread can
    also be forced to check the queue by setting the :func:`flush_notification` event.

    - If no items are found, the thread will go back to sleep.
    - If items are found, the worker thread will send items to the specified service in batches of :func:`send_buffer_size`.

    If no queue items are found for :func:`send_time` seconds, the worker thread will shut down (and :func:`start` will
    need  to be called again).
    """

    def __init__(self, service_endpoint_uri=None):
        """Initializes a new instance of the class.

        Args:
            sender (String) service_endpoint_uri the address of the service to send telemetry data to.
        """
        self._send_interval = 1.0
        self._send_remaining_time = 0
        self._send_time = 3.0
        self._lock_send_remaining_time = Lock()
        SenderBase.__init__(self, service_endpoint_uri or DEFAULT_ENDPOINT_URL)

    @property
    def send_interval(self):
        """The time span in seconds at which the the worker thread will check the :func:`queue` for items (defaults to: 1.0).

        Args:
            value (int) the interval in seconds.

        Returns:
            int. the interval in seconds.
        """
        return self._send_interval

    @send_interval.setter
    def send_interval(self, value):
        """The time span in seconds at which the the worker thread will check the :func:`queue` for items (defaults to: 1.0).

        Args:
            value (int) the interval in seconds.

        Returns:
            int. the interval in seconds.
        """
        self._send_interval = value

    @property
    def send_time(self):
        """The time span in seconds at which the the worker thread will check the :func:`queue` for items (defaults to: 1.0).

        Args:
            value (int) the interval in seconds.

        Returns:
            int. the interval in seconds.
        """
        return self._send_time

    @send_time.setter
    def send_time(self, value):
        """The time span in seconds at which the the worker thread will check the :func:`queue` for items (defaults to: 1.0).

        Args:
            value (int) the interval in seconds.

        Returns:
            int. the interval in seconds.
        """
        self._send_time = value

    def start(self):
        """Starts a new sender thread if none is not already there
        """
        with self._lock_send_remaining_time:
            if self._send_remaining_time <= 0.0:
                local_send_interval = self._send_interval
                if self._send_interval < 0.1:
                    local_send_interval = 0.1
                self._send_remaining_time = self._send_time
                if self._send_remaining_time < local_send_interval:
                    self._send_remaining_time = local_send_interval
                thread = Thread(target=self._run)
                thread.daemon = True
                thread.start()

    def stop(self):
        """Gracefully stops the sender thread if one is there.
        """
        with self._lock_send_remaining_time:
            self._send_remaining_time = 0.0

    def _run(self):
        # save the queue locally
        local_queue = self._queue
        if not local_queue:
            self.stop()
            return

        # fix up the send interval (can't be lower than 100ms)
        local_send_interval = self._send_interval
        if self._send_interval < 0.1:
            local_send_interval = 0.1
        local_send_time = self._send_time
        if local_send_time < local_send_interval:
            local_send_time = local_send_interval
        while True:
            while True:
                # get at most send_buffer_size items from the queue
                counter = self._send_buffer_size
                data = []
                while counter > 0:
                    item = local_queue.get()
                    if not item:
                        break
                    data.append(item)
                    counter -= 1

                # if we didn't get any items from the queue, we're done here
                if len(data) == 0:
                    break

                # reset the send time
                with self._lock_send_remaining_time:
                    self._send_remaining_time = local_send_time

                # finally send the data
                self.send(data)

            # wait at most send_interval (or until we get signalled)
            result = local_queue.flush_notification.wait(local_send_interval)
            if result:
                local_queue.flush_notification.clear()
                continue

            # decrement the remaining time
            local_remaining_time = 0
            with self._lock_send_remaining_time:
                self._send_remaining_time -= local_send_interval
                local_remaining_time = self._send_remaining_time

            if local_remaining_time <= 0:
                break



class SeverityLevel(object):
    """Data contract class for type SeverityLevel."""
    # Enumeration value verbose
    verbose = 0

    # Enumeration value information
    information = 1

    # Enumeration value warning
    warning = 2

    # Enumeration value error
    error = 3

    # Enumeration value critical
    critical = 4



import collections
import copy
from .Utils import _write_complex_object


class Domain(object):
    """Data contract class for type Domain.
    """
    _defaults = collections.OrderedDict([
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
        }
        self._initialize()

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



from .SenderBase import SenderBase, DEFAULT_ENDPOINT_URL


class SynchronousSender(SenderBase):
    """A synchronous sender that works in conjunction with the :class:`SynchronousQueue`. The queue will call
    :func:`send` on the current instance with the data to send.
    """

    def __init__(self, service_endpoint_uri=None):
        """Initializes a new instance of the class.

        Args:
            sender (String) service_endpoint_uri the address of the service to send telemetry data to.
        """
        SenderBase.__init__(self, service_endpoint_uri or DEFAULT_ENDPOINT_URL)



from .middleware import ApplicationInsightsMiddleware
from .logging import LoggingHandler
from . import common

__all__ = ['ApplicationInsightsMiddleware', 'LoggingHandler', 'create_client']


def create_client():
    """Returns an :class:`applicationinsights.TelemetryClient` instance using the instrumentation key
    and other settings found in the current Django project's `settings.py` file."""
    return common.create_client()



from os import getenv
import platform

try:
    from werkzeug.exceptions import HTTPException
except ImportError:
    HTTPException = None

from applicationinsights import TelemetryClient
from applicationinsights.channel import AsynchronousSender
from applicationinsights.channel import AsynchronousQueue
from applicationinsights.channel import TelemetryChannel
from applicationinsights.logging import LoggingHandler
from applicationinsights.requests import WSGIApplication

CONF_PREFIX = "APPINSIGHTS"

CONF_KEY = CONF_PREFIX + "_INSTRUMENTATIONKEY"
CONF_ENDPOINT_URI = CONF_PREFIX + "_ENDPOINT_URI"
CONF_DISABLE_REQUEST_LOGGING = CONF_PREFIX + "_DISABLE_REQUEST_LOGGING"
CONF_DISABLE_TRACE_LOGGING = CONF_PREFIX + "_DISABLE_TRACE_LOGGING"
CONF_DISABLE_EXCEPTION_LOGGING = CONF_PREFIX + "_DISABLE_EXCEPTION_LOGGING"


class AppInsights(object):
    """ This class represents a Flask extension that enables request telemetry,
    trace logging and exception logging for a Flask application. The telemetry
    will be sent to Application Insights service using the supplied
    instrumentation key.

    The following Flask config variables can be used to configure the extension:

    - Set ``APPINSIGHTS_INSTRUMENTATIONKEY`` to a string to provide the
      instrumentation key to send telemetry to application insights.
      Alternatively, this value can also be provided via an environment variable
      of the same name.

    - Set ``APPINSIGHTS_ENDPOINT_URI`` to a string to customize the telemetry
      endpoint to which Application Insights will send the telemetry.

    - Set ``APPINSIGHTS_DISABLE_REQUEST_LOGGING`` to ``False`` to disable
      logging of Flask requests to Application Insights.

    - Set ``APPINSIGHTS_DISABLE_TRACE_LOGGING`` to ``False`` to disable logging
      of all log traces to Application Insights.

    - Set ``APPINSIGHTS_DISABLE_EXCEPTION_LOGGING`` to ``False`` to disable
      logging of all exceptions to Application Insights.

    .. code:: python

            from flask import Flask
            from applicationinsights.flask.ext import AppInsights

            # instantiate the Flask application
            app = Flask(__name__)
            app.config['APPINSIGHTS_INSTRUMENTATIONKEY'] = '<YOUR INSTRUMENTATION KEY GOES HERE>'

            # log requests, traces and exceptions to the Application Insights service
            appinsights = AppInsights(app)

            # define a simple route
            @app.route('/')
            def hello_world():
                return 'Hello World!'

            # run the application
            if __name__ == '__main__':
                app.run()
    """

    def __init__(self, app=None):
        """
        Initialize a new instance of the extension.

        Args:
            app (flask.Flask). the Flask application for which to initialize the extension.
        """
        self._key = None
        self._endpoint_uri = None
        self._channel = None
        self._requests_middleware = None
        self._trace_log_handler = None
        self._exception_telemetry_client = None

        if app:
            self.init_app(app)

    def init_app(self, app):
        """
        Initializes the extension for the provided Flask application.

        Args:
            app (flask.Flask). the Flask application for which to initialize the extension.
        """
        self._key = app.config.get(CONF_KEY) or getenv(CONF_KEY)

        if not self._key:
            return

        self._endpoint_uri = app.config.get(CONF_ENDPOINT_URI)
        sender = AsynchronousSender(self._endpoint_uri)

        queue = AsynchronousQueue(sender)
        self._channel = TelemetryChannel(None, queue)

        self.context.cloud.role = app.name
        self.context.cloud.role_instance = platform.node()

        self._init_request_logging(app)
        self._init_trace_logging(app)
        self._init_exception_logging(app)

    @property
    def context(self):
        """
        Accesses the telemetry context.

        Returns:
            (applicationinsights.channel.TelemetryContext). The Application Insights telemetry context.
        """
        return self._channel.context

    def _init_request_logging(self, app):
        """
        Sets up request logging unless ``APPINSIGHTS_DISABLE_REQUEST_LOGGING``
        is set in the Flask config.

        Args:
            app (flask.Flask). the Flask application for which to initialize the extension.
        """
        enabled = not app.config.get(CONF_DISABLE_REQUEST_LOGGING, False)

        if not enabled:
            return

        self._requests_middleware = WSGIApplication(
            self._key, app.wsgi_app, telemetry_channel=self._channel)

        app.wsgi_app = self._requests_middleware

    def _init_trace_logging(self, app):
        """
        Sets up trace logging unless ``APPINSIGHTS_DISABLE_TRACE_LOGGING`` is
        set in the Flask config.

        Args:
            app (flask.Flask). the Flask application for which to initialize the extension.
        """
        enabled = not app.config.get(CONF_DISABLE_TRACE_LOGGING, False)

        if not enabled:
            return

        self._trace_log_handler = LoggingHandler(
            self._key, telemetry_channel=self._channel)

        app.logger.addHandler(self._trace_log_handler)

    def _init_exception_logging(self, app):
        """
        Sets up exception logging unless ``APPINSIGHTS_DISABLE_EXCEPTION_LOGGING``
        is set in the Flask config.

        Args:
            app (flask.Flask). the Flask application for which to initialize the extension.
        """
        enabled = not app.config.get(CONF_DISABLE_EXCEPTION_LOGGING, False)

        if not enabled:
            return

        exception_telemetry_client = TelemetryClient(
            self._key, telemetry_channel=self._channel)

        @app.errorhandler(Exception)
        def exception_handler(exception):
            if HTTPException and isinstance(exception, HTTPException):
                return exception

            try:
                raise exception
            except Exception:
                exception_telemetry_client.track_exception()
            finally:
                raise exception

        self._exception_telemetry_client = exception_telemetry_client

    def flush(self):
        """Flushes the queued up telemetry to the service.
        """
        if self._requests_middleware:
            self._requests_middleware.flush()

        if self._trace_log_handler:
            self._trace_log_handler.flush()

        if self._exception_telemetry_client:
            self._exception_telemetry_client.flush()



import collections
import copy
import logging
from .Utils import _write_complex_object


class MessageData(object):
    """Data contract class for type MessageData.
    """

    ENVELOPE_TYPE_NAME = 'Microsoft.ApplicationInsights.Message'

    DATA_TYPE_NAME = 'MessageData'

    PYTHON_LOGGING_LEVELS = {
        'DEBUG': 0,
        'INFO': 1,
        'WARNING': 2,
        'ERROR': 3,
        'CRITICAL': 4,
        logging.DEBUG: 0,
        logging.INFO: 1,
        logging.WARNING: 2,
        logging.ERROR: 3,
        logging.CRITICAL: 4
    }

    _defaults = collections.OrderedDict([
        ('ver', 2),
        ('message', None),
        ('severityLevel', None),
        ('properties', {})
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
            'ver': 2,
            'message': None,
        }
        self._initialize()

    @property
    def ver(self):
        """The ver property.

        Returns:
            (int). the property value. (defaults to: 2)
        """
        return self._values['ver']

    @ver.setter
    def ver(self, value):
        """The ver property.

        Args:
            value (int). the property value.
        """
        self._values['ver'] = value

    @property
    def message(self):
        """The message property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        return self._values['message']

    @message.setter
    def message(self, value):
        """The message property.

        Args:
            value (string). the property value.
        """
        self._values['message'] = value

    @property
    def severity_level(self):
        """The severity_level property.

        Returns:
            (int). the property value. (defaults to: None)
        """
        if 'severityLevel' in self._values:
            return self._values['severityLevel']
        return self._defaults['severityLevel']

    @severity_level.setter
    def severity_level(self, value):
        """The severity_level property.

        Args:
            value (int). the property value.
        """
        if value == self._defaults['severityLevel'] and 'severityLevel' in self._values:
            del self._values['severityLevel']
        else:
            self._values['severityLevel'] = value

    @property
    def properties(self):
        """The properties property.

        Returns:
            (hash). the property value. (defaults to: {})
        """
        if 'properties' in self._values:
            return self._values['properties']
        self._values['properties'] = copy.deepcopy(
            self._defaults['properties'])
        return self._values['properties']

    @properties.setter
    def properties(self, value):
        """The properties property.

        Args:
            value (hash). the property value.
        """
        if value == self._defaults['properties'] and 'properties' in self._values:
            del self._values['properties']
        else:
            self._values['properties'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



from setuptools import setup, find_packages  # Always prefer setuptools over distutils
from codecs import open  # To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='appinsights',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # http://packaging.python.org/en/latest/tutorial.html#version
    version='0.11.14',

    description='This project extends the Application Insights API surface to support Python.',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/brennerm/python-appinsights',
    download_url='https://github.com/brennerm/python-appinsights',

    # Author details
    author='Max Brenner',
    author_email='xamrennerb@gmail.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',

        # operating systems
        'Operating System :: OS Independent',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    # What does your project relate to?
    keywords='analytics applicationinsights telemetry appinsights development',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),

    test_suite='tests.applicationinsights_tests'
)





from . import common
from applicationinsights import logging

import sys


class LoggingHandler(logging.LoggingHandler):
    """This class is a LoggingHandler that uses the same settings as the Django middleware to configure
    the telemetry client.  This can be referenced from LOGGING in your Django settings.py file.  As an
    example, this code would send all Django log messages--WARNING and up--to Application Insights:

    .. code:: python

            LOGGING = {
                'version': 1,
                'disable_existing_loggers': False,
                'handlers': {
                    # The application insights handler is here
                    'appinsights': {
                        'class': 'applicationinsights.django.LoggingHandler',
                        'level': 'WARNING'
                    }
                },
                'loggers': {
                    'django': {
                        'handlers': ['appinsights'],
                        'level': 'WARNING',
                        'propagate': True,
                    }
                }
            }

            # You will need this anyway if you're using the middleware.
            # See the middleware documentation for more information on configuring
            # this setting:
            APPLICATION_INSIGHTS = {
                'ikey': '00000000-0000-0000-0000-000000000000'
            }
    """

    def __init__(self, *args, **kwargs):
        client = common.create_client()
        new_kwargs = {}
        new_kwargs.update(kwargs)
        new_kwargs['telemetry_channel'] = client.channel
        super(LoggingHandler, self).__init__(
            client.context.instrumentation_key, *args, **new_kwargs)



import collections
import copy
from .Utils import _write_complex_object


class Data(object):
    """Data contract class for type Data.
    """
    _defaults = collections.OrderedDict([
        ('baseType', None),
        ('baseData', None)
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
            'baseData': None
        }
        self._initialize()

    @property
    def base_type(self):
        """The base_type property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'baseType' in self._values:
            return self._values['baseType']
        return self._defaults['baseType']

    @base_type.setter
    def base_type(self, value):
        """The base_type property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['baseType'] and 'baseType' in self._values:
            del self._values['baseType']
        else:
            self._values['baseType'] = value

    @property
    def base_data(self):
        """The base_data property.

        Returns:
            (object). the property value. (defaults to: None)
        """
        return self._values['baseData']

    @base_data.setter
    def base_data(self, value):
        """The base_data property.

        Args:
            value (object). the property value.
        """
        self._values['baseData'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



import collections
import copy
from .Utils import _write_complex_object


class Internal(object):
    """Data contract class for type Internal.
    """
    _defaults = collections.OrderedDict([
        ('ai.internal.sdkVersion', None),
        ('ai.internal.agentVersion', None),
        ('ai.internal.nodeName', None)
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
        }
        self._initialize()

    @property
    def sdk_version(self):
        """The sdk_version property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.internal.sdkVersion' in self._values:
            return self._values['ai.internal.sdkVersion']
        return self._defaults['ai.internal.sdkVersion']

    @sdk_version.setter
    def sdk_version(self, value):
        """The sdk_version property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.internal.sdkVersion'] and 'ai.internal.sdkVersion' in self._values:
            del self._values['ai.internal.sdkVersion']
        else:
            self._values['ai.internal.sdkVersion'] = value

    @property
    def agent_version(self):
        """The agent_version property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.internal.agentVersion' in self._values:
            return self._values['ai.internal.agentVersion']
        return self._defaults['ai.internal.agentVersion']

    @agent_version.setter
    def agent_version(self, value):
        """The agent_version property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.internal.agentVersion'] and 'ai.internal.agentVersion' in self._values:
            del self._values['ai.internal.agentVersion']
        else:
            self._values['ai.internal.agentVersion'] = value

    @property
    def node_name(self):
        """The node_name property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'ai.internal.nodeName' in self._values:
            return self._values['ai.internal.nodeName']
        return self._defaults['ai.internal.nodeName']

    @node_name.setter
    def node_name(self, value):
        """The node_name property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['ai.internal.nodeName'] and 'ai.internal.nodeName' in self._values:
            del self._values['ai.internal.nodeName']
        else:
            self._values['ai.internal.nodeName'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)



import collections
import copy
from .Utils import _write_complex_object


class RemoteDependencyData(object):
    """Data contract class for type RemoteDependencyData.
    """
    ENVELOPE_TYPE_NAME = 'Microsoft.ApplicationInsights.RemoteDependency'

    DATA_TYPE_NAME = 'RemoteDependencyData'

    _defaults = collections.OrderedDict([
        ('ver', 2),
        ('name', None),
        ('id', None),
        ('resultCode', None),
        ('duration', None),
        ('success', True),
        ('data', None),
        ('target', None),
        ('type', None),
        ('properties', {}),
        ('measurements', {})
    ])

    def __init__(self):
        """Initializes a new instance of the class.
        """
        self._values = {
            'ver': 2,
            'name': None,
            'duration': None,
            'success': True,
        }
        self._initialize()

    @property
    def ver(self):
        """The ver property.

        Returns:
            (int). the property value. (defaults to: 2)
        """
        return self._values['ver']

    @ver.setter
    def ver(self, value):
        """The ver property.

        Args:
            value (int). the property value.
        """
        self._values['ver'] = value

    @property
    def name(self):
        """The name property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        return self._values['name']

    @name.setter
    def name(self, value):
        """The name property.

        Args:
            value (string). the property value.
        """
        self._values['name'] = value

    @property
    def id(self):
        """The id property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'id' in self._values:
            return self._values['id']
        return self._defaults['id']

    @id.setter
    def id(self, value):
        """The id property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['id'] and 'id' in self._values:
            del self._values['id']
        else:
            self._values['id'] = value

    @property
    def result_code(self):
        """The result_code property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'resultCode' in self._values:
            return self._values['resultCode']
        return self._defaults['resultCode']

    @result_code.setter
    def result_code(self, value):
        """The result_code property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['resultCode'] and 'resultCode' in self._values:
            del self._values['resultCode']
        else:
            self._values['resultCode'] = value

    @property
    def duration(self):
        """The duration property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        return self._values['duration']

    @duration.setter
    def duration(self, value):
        """The duration property.

        Args:
            value (string). the property value.
        """
        self._values['duration'] = value

    @property
    def success(self):
        """The success property.

        Returns:
            (bool). the property value. (defaults to: True)
        """
        if 'success' in self._values:
            return self._values['success']
        return self._defaults['success']

    @success.setter
    def success(self, value):
        """The success property.

        Args:
            value (bool). the property value.
        """
        if value == self._defaults['success'] and 'success' in self._values:
            del self._values['success']
        else:
            self._values['success'] = value

    @property
    def data(self):
        """The data property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'data' in self._values:
            return self._values['data']
        return self._defaults['data']

    @data.setter
    def data(self, value):
        """The data property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['data'] and 'data' in self._values:
            del self._values['data']
        else:
            self._values['data'] = value

    @property
    def target(self):
        """The target property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'target' in self._values:
            return self._values['target']
        return self._defaults['target']

    @target.setter
    def target(self, value):
        """The target property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['target'] and 'target' in self._values:
            del self._values['target']
        else:
            self._values['target'] = value

    @property
    def type(self):
        """The type property.

        Returns:
            (string). the property value. (defaults to: None)
        """
        if 'type' in self._values:
            return self._values['type']
        return self._defaults['type']

    @type.setter
    def type(self, value):
        """The type property.

        Args:
            value (string). the property value.
        """
        if value == self._defaults['type'] and 'type' in self._values:
            del self._values['type']
        else:
            self._values['type'] = value

    @property
    def properties(self):
        """The properties property.

        Returns:
            (hash). the property value. (defaults to: {})
        """
        if 'properties' in self._values:
            return self._values['properties']
        self._values['properties'] = copy.deepcopy(
            self._defaults['properties'])
        return self._values['properties']

    @properties.setter
    def properties(self, value):
        """The properties property.

        Args:
            value (hash). the property value.
        """
        if value == self._defaults['properties'] and 'properties' in self._values:
            del self._values['properties']
        else:
            self._values['properties'] = value

    @property
    def measurements(self):
        """The measurements property.

        Returns:
            (hash). the property value. (defaults to: {})
        """
        if 'measurements' in self._values:
            return self._values['measurements']
        self._values['measurements'] = copy.deepcopy(
            self._defaults['measurements'])
        return self._values['measurements']

    @measurements.setter
    def measurements(self, value):
        """The measurements property.

        Args:
            value (hash). the property value.
        """
        if value == self._defaults['measurements'] and 'measurements' in self._values:
            del self._values['measurements']
        else:
            self._values['measurements'] = value

    def _initialize(self):
        """Initializes the current instance of the object.
        """
        pass

    def write(self):
        """Writes the contents of this object and returns the content as a dict object.

        Returns:
            (dict). the object that represents the same data as the current instance.
        """
        return _write_complex_object(self._defaults, self._values)
